/*  BMP - Cross-platform multimedia player
 *  Copyright (C) 2003-2004  Edward Brocklesby, Chong Kai Xiong,
 *            Milosz Derezynski, David Lau, Michiel sikkes.
 *
 *  Based on XMMS:
 *  Copyright (C) 1998-2003  XMMS development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "main.h"

#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <gdk/gdk.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <ctype.h>
#include <time.h>

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "libbeep/configdb.h"
#include "libbeep/beepctrl.h"
#include "libbeep/util.h"

#include "controlsocket.h"
#include "dnd.h"
#include "dock.h"
#include "effect.h"
#include "equalizer.h"
#include "general.h"
#include "hints.h"
#include "input.h"
#include "logger.h"
#include "mainwin.h"
#include "output.h"
#include "playlist.h"
#include "playlistwin.h"
#include "pluginenum.h"
#include "prefswin.h"
#include "skin.h"
#include "skinwin.h"
#include "util.h"
#include "visualization.h"

#include "pixmaps.h"
#include "beep_mini.xpm"


/* Translatable string for beep.desktop's comment field */
const gchar *desktop_comment = N_("Beep Media Player");

const gchar *application_name = N_("Beep Media Player");


struct _BmpCmdLineOpt {
    GList *filenames;
    gint session;
    gboolean play, stop, pause, fwd, rew, play_pause, playcd;
    gboolean enqueue, mainwin, remote;
    gboolean load_skins;
    gchar *previous_session_id;

    gboolean xmms_compat_mode;
};

typedef struct _BmpCmdLineOpt BmpCmdLineOpt;


BmpConfig cfg;

BmpConfig bmp_default_config = {
    MAINWIN_DEFAULT_POS_X,      /* mainwin x position */
    MAINWIN_DEFAULT_POS_Y,      /* mainwin y position */
    EQUALIZER_DEFAULT_POS_X,    /* equalizer x position */
    EQUALIZER_DEFAULT_POS_Y,    /* equalizer y position */
    PLAYLISTWIN_DEFAULT_POS_X,  /* playlistwin x position */
    PLAYLISTWIN_DEFAULT_POS_Y,  /* playlistwin y position */
    PLAYLISTWIN_DEFAULT_WIDTH,  /* playlistwin width */
    PLAYLISTWIN_DEFAULT_HEIGHT, /* playlistwin height */
    10,                         /* snap distance */
    FALSE,                      /* real-time priority */
    FALSE, FALSE,               /* shuffle, repeat */
    FALSE, TRUE,                /* double size, autoscroll */
    TRUE,                       /* analyzer peaks */
    FALSE,                      /* equalizer autoload */
    FALSE,                      /* easy move */
    FALSE,                      /* equalizer active */
    FALSE,                      /* playlistwin visible */
    FALSE,                      /* equalizer visible */
    TRUE,                       /* player visible */
    FALSE,                      /* player shaded */
    FALSE,                      /* playlistwin shaded */
    FALSE,                      /* equalizer shaded */
    FALSE,                      /* allow multiple instances */
    TRUE,                       /* always show cb */
    TRUE, TRUE,                 /* convert '_' and %20 */
    TRUE,                       /* show numbers in playlist */
    TRUE,                       /* snap windows */
    TRUE,                       /* save window positions */
    TRUE,                       /* dim titlebar */
    TRUE,                       /* get playlist info on load */
    TRUE,                       /* get playlist info on demand */
    TRUE,                       /* equalizer doublesize linked */
    FALSE,                      /* sort jump to file */
    FALSE,                      /* use effect plugins */
    FALSE,                      /* always on top */
    FALSE,                      /* sticky */
    FALSE,                      /* no playlist advance */
    TRUE,                       /* smooth title scrolling */
    TRUE,                       /* use playlist metadata */
    TRUE,                       /* warn about unplayables */
    FALSE,                      /* use \ as directory delimiter */
    FALSE,                      /* random skin on play */
    FALSE,                      /* use fontsets */
    FALSE,                      /* use X font for mainwin */
    TRUE,			/* use custom cursors */
    0.0,                        /* equalizer preamp */
    {0, 0, 0, 0, 0,             /* equalizer bands */
     0, 0, 0, 0, 0},
    NULL,                       /* skin */
    NULL,                       /* output plugin */
    NULL,                       /* file selector path */
    NULL,                       /* playlist path */
    NULL,                       /* playlist font */
    NULL,                       /* mainwin font */
    NULL,                       /* disabled input plugins */
    NULL,                       /* enabled general plugins */
    NULL,                       /* enabled visualization plugins */
    NULL,                       /* enabled effect plugins */
    NULL,                       /* equalizer preset default file */
    NULL,                       /* equalizer preset extension */
    NULL,                       /* URL history */
    0,                          /* timer mode */
    VIS_ANALYZER,               /* visualizer type */
    ANALYZER_NORMAL,            /* analyzer mode */
    ANALYZER_BARS,              /* analyzer type */
    SCOPE_DOT,                  /* scope mode */
    VU_SMOOTH,                  /* VU mode */
    REFRESH_FULL,               /* visualizer refresh rate */
    FALLOFF_FAST,               /* analyzer fall off rate */
    FALLOFF_SLOW,               /* peaks fall off rate */
    0,                          /* playlist position */
    2,                          /* pause between songs time */
    FALSE,                      /* pause between songs */
    2,                          /* 2nd preset (ARTIST - TITLE) */
    FALSE,                      /* show window decorations */
    8,                          /* mouse wheel scroll step */
    FALSE,                      /* playlist transparent */
    NULL,                       /* title format */
    FALSE,                      /* software volume control enabled */
    TRUE,                       /* XMMS compatibility mode */
    TRUE,                       /* extra eq filtering */
    3,                          /* scroll pl by */
};

typedef struct bmp_cfg_boolent_t {
    char const *be_vname;
    gboolean *be_vloc;
    gboolean be_wrt;
} bmp_cfg_boolent;

typedef struct bmp_cfg_nument_t {
    char const *ie_vname;
    gint *ie_vloc;
    gboolean ie_wrt;
} bmp_cfg_nument;

typedef struct bmp_cfg_strent_t {
    char const *se_vname;
    char **se_vloc;
    gboolean se_wrt;
} bmp_cfg_strent;

static const gchar *bmp_titlestring_presets[] = {
    "%t",
    "%p - %t",
    "%p - %a - %t",
    "%a - %t",
};

static bmp_cfg_boolent bmp_boolents[] = {
    {"allow_multiple_instances", &cfg.allow_multiple_instances, TRUE},
    {"use_realtime", &cfg.use_realtime, TRUE},
    {"always_show_cb", &cfg.always_show_cb, TRUE},
    {"convert_underscore", &cfg.convert_underscore, TRUE},
    {"convert_twenty", &cfg.convert_twenty, TRUE},
    {"show_numbers_in_pl", &cfg.show_numbers_in_pl, TRUE},
    {"snap_windows", &cfg.snap_windows, TRUE},
    {"save_window_positions", &cfg.save_window_position, TRUE},
    {"dim_titlebar", &cfg.dim_titlebar, TRUE},
    {"get_info_on_load", &cfg.get_info_on_load, TRUE},
    {"get_info_on_demand", &cfg.get_info_on_demand, TRUE},
    {"eq_doublesize_linked", &cfg.eq_doublesize_linked, TRUE},
    {"no_playlist_advance", &cfg.no_playlist_advance, TRUE},
    {"sort_jump_to_file", &cfg.sort_jump_to_file, TRUE},
    {"use_pl_metadata", &cfg.use_pl_metadata, TRUE},
    {"warn_about_unplayables", &cfg.warn_about_unplayables, TRUE},
    {"smooth_title_scroll", &cfg.smooth_title_scroll, TRUE},
    {"use_backslash_as_dir_delimiter", &cfg.use_backslash_as_dir_delimiter,
     TRUE},
    {"player_shaded", &cfg.player_shaded, TRUE},
    {"player_visible", &cfg.player_visible, TRUE},
    {"shuffle", &cfg.shuffle, TRUE},
    {"repeat", &cfg.repeat, TRUE},
    {"doublesize", &cfg.doublesize, TRUE},
    {"autoscroll_songname", &cfg.autoscroll, TRUE},
    {"playlist_shaded", &cfg.playlist_shaded, TRUE},
    {"playlist_visible", &cfg.playlist_visible, TRUE},
    {"playlist_transparent", &cfg.playlist_transparent, TRUE},
    {"use_fontsets", &cfg.use_fontsets, TRUE},
    {"mainwin_use_xfont", &cfg.mainwin_use_xfont, TRUE},
    {"equalizer_visible", &cfg.equalizer_visible, TRUE},
    {"equalizer_active", &cfg.equalizer_active, TRUE},
    {"equalizer_shaded", &cfg.equalizer_shaded, TRUE},
    {"equalizer_autoload", &cfg.equalizer_autoload, TRUE},
    {"easy_move", &cfg.easy_move, TRUE},
    {"use_eplugins", &cfg.use_eplugins, TRUE},
    {"always_on_top", &cfg.always_on_top, TRUE},
    {"sticky", &cfg.sticky, TRUE},
    {"random_skin_on_play", &cfg.random_skin_on_play, TRUE},
    {"pause_between_songs", &cfg.pause_between_songs, TRUE},
    {"show_wm_decorations", &cfg.show_wm_decorations, TRUE},
    {"xmms_compat_mode", &cfg.xmms_compat_mode, TRUE},
    {"eq_extra_filtering", &cfg.eq_extra_filtering, TRUE},
    {"analyzer_peaks", &cfg.analyzer_peaks, TRUE},
    {"custom_cursors", &cfg.custom_cursors, TRUE}
};

static gint ncfgbent = G_N_ELEMENTS(bmp_boolents);

static bmp_cfg_nument bmp_numents[] = {
    {"player_x", &cfg.player_x, TRUE},
    {"player_y", &cfg.player_y, TRUE},
    {"timer_mode", &cfg.timer_mode, TRUE},
    {"vis_type", &cfg.vis_type, TRUE},
    {"analyzer_mode", &cfg.analyzer_mode, TRUE},
    {"analyzer_type", &cfg.analyzer_type, TRUE},
    {"scope_mode", &cfg.scope_mode, TRUE},
    {"vu_mode", &cfg.vu_mode, TRUE},
    {"vis_refresh_rate", &cfg.vis_refresh, TRUE},
    {"analyzer_falloff", &cfg.analyzer_falloff, TRUE},
    {"peaks_falloff", &cfg.peaks_falloff, TRUE},
    {"playlist_x", &cfg.playlist_x, TRUE},
    {"playlist_y", &cfg.playlist_y, TRUE},
    {"playlist_width", &cfg.playlist_width, TRUE},
    {"playlist_height", &cfg.playlist_height, TRUE},
    {"playlist_position", &cfg.playlist_position, TRUE},
    {"equalizer_x", &cfg.equalizer_x, TRUE},
    {"equalizer_y", &cfg.equalizer_y, TRUE},
    {"snap_distance", &cfg.snap_distance, TRUE},
    {"pause_between_songs_time", &cfg.pause_between_songs_time, TRUE},
    {"mouse_wheel_change", &cfg.mouse_change, TRUE},
    {"scroll_pl_by", &cfg.scroll_pl_by, TRUE},
    {"titlestring_preset", &cfg.titlestring_preset, TRUE},
};

static gint ncfgient = G_N_ELEMENTS(bmp_numents);

static bmp_cfg_strent bmp_strents[] = {
    {"playlist_font", &cfg.playlist_font, TRUE},
    {"mainwin_font", &cfg.mainwin_font, TRUE},
    {"eqpreset_default_file", &cfg.eqpreset_default_file, TRUE},
    {"eqpreset_extension", &cfg.eqpreset_extension, TRUE},
    {"skin", &cfg.skin, FALSE},
    {"output_plugin", &cfg.outputplugin, FALSE},
    {"enabled_gplugins", &cfg.enabled_gplugins, FALSE},
    {"enabled_vplugins", &cfg.enabled_vplugins, FALSE},
    {"enabled_eplugins", &cfg.enabled_eplugins, FALSE},
    {"filesel_path", &cfg.filesel_path, FALSE},
    {"playlist_path", &cfg.playlist_path, FALSE},
    {"disabled_iplugins", &cfg.disabled_iplugins, TRUE},
    {"generic_title_format", &cfg.gentitle_format, TRUE},
};

static gint ncfgsent = G_N_ELEMENTS(bmp_strents);

gchar *bmp_user_dir = NULL;
gchar *bmp_user_plugin_dir = NULL;
gchar *bmp_user_skin_dir = NULL;
gchar *bmp_user_skin_thumb_dir = NULL;
gchar *bmp_config_filename = NULL;
gchar *bmp_playlist_filename = NULL;
gchar *bmp_log_filename = NULL;

gchar *disabled_iplugins = NULL;

GList *dock_window_list = NULL;

gboolean pposition_broken = FALSE;


static void
dump_version(void)
{
    g_printf("%s %s\n", PACKAGE, VERSION);
}

/* FIXME: this function should just return cfg.gentitle_format,
   this special handling of presets will introduce a couple of
   other headaches */
const gchar *
xmms_get_gentitle_format(void)
{
    gint titlestring_preset = cfg.titlestring_preset;
    if (titlestring_preset != 4) {
	return bmp_titlestring_presets[titlestring_preset];
    } else {
    	return cfg.gentitle_format;
    }
}

static void
make_directory(const gchar * path, mode_t mode)
{
    if (mkdir(path, mode) == 0)
        return;

    if (errno == EEXIST)
        return;

    g_printerr(_("Could not create directory (%s): %s"), path,
               g_strerror(errno));
}

static void
bmp_make_user_dir(void)
{
    const mode_t mode755 = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

    make_directory(bmp_user_dir, mode755);
    make_directory(bmp_user_skin_dir, mode755);
    make_directory(bmp_user_skin_thumb_dir, mode755);
    make_directory(bmp_user_plugin_dir, mode755);
}

static void
bmp_free_paths(void)
{
    g_free(bmp_user_dir);
    g_free(bmp_user_plugin_dir);
    g_free(bmp_user_skin_dir);
    g_free(bmp_user_skin_thumb_dir);
    g_free(bmp_config_filename);
    g_free(bmp_playlist_filename);
    g_free(bmp_log_filename);
}

static void
bmp_init_paths(void)
{
    bmp_user_dir = g_build_filename(g_get_home_dir(), BMP_RCPATH, NULL);

    bmp_user_plugin_dir =
        g_build_filename(bmp_user_dir, BMP_USER_PLUGIN_DIR_BASENAME, NULL);
    bmp_user_skin_dir =
        g_build_filename(bmp_user_dir, BMP_SKIN_DIR_BASENAME, NULL);
    bmp_user_skin_thumb_dir =
        g_build_filename(bmp_user_skin_dir, BMP_SKIN_THUMB_DIR_BASENAME,
                         NULL);
    bmp_config_filename =
        g_build_filename(bmp_user_dir, BMP_CONFIG_BASENAME, NULL);
    bmp_playlist_filename =
        g_build_filename(bmp_user_dir, BMP_PLAYLIST_BASENAME, NULL);
    bmp_log_filename = g_build_filename(bmp_user_dir, BMP_LOG_BASENAME, NULL);

    g_atexit(bmp_free_paths);
}

void
bmp_config_load(void)
{
    ConfigDb *db;
    gint i, length;

    memcpy(&cfg, &bmp_default_config, sizeof(BmpConfig));

    db = bmp_cfg_db_open();
    for (i = 0; i < ncfgbent; ++i) {
        bmp_cfg_db_get_bool(db, NULL,
                            bmp_boolents[i].be_vname,
                            bmp_boolents[i].be_vloc);
    }

    for (i = 0; i < ncfgient; ++i) {
        bmp_cfg_db_get_int(db, NULL,
                           bmp_numents[i].ie_vname,
                           bmp_numents[i].ie_vloc);
    }

    for (i = 0; i < ncfgsent; ++i) {
        bmp_cfg_db_get_string(db, NULL,
                              bmp_strents[i].se_vname,
                              bmp_strents[i].se_vloc);
    }

    bmp_cfg_db_get_float(db, NULL, "equalizer_preamp", &cfg.equalizer_preamp);
    for (i = 0; i < 10; i++) {
        gchar eqtext[18];

        g_snprintf(eqtext, sizeof(eqtext), "equalizer_band%d", i);
        bmp_cfg_db_get_float(db, NULL, eqtext, &cfg.equalizer_bands[i]);
    }
    if (bmp_cfg_db_get_int(db, NULL, "url_history_length", &length)) {
        for (i = 1; i <= length; i++) {
            gchar str[19], *tmp;

            g_snprintf(str, sizeof(str), "url_history%d", i);
            if (bmp_cfg_db_get_string(db, NULL, str, &tmp))
                cfg.url_history = g_list_append(cfg.url_history, tmp);
        }
    }

    bmp_cfg_db_close(db);


    if (cfg.playlist_font && strlen(cfg.playlist_font) == 0) {
        g_free(cfg.playlist_font);
        cfg.playlist_font = NULL;
    }

    if (cfg.mainwin_font && strlen(cfg.mainwin_font) == 0) {
        g_free(cfg.mainwin_font);
        cfg.mainwin_font = NULL;
    }

    if (!cfg.playlist_font)
        cfg.playlist_font = g_strdup(PLAYLISTWIN_DEFAULT_FONT);

    if (!cfg.mainwin_font)
        cfg.mainwin_font = g_strdup(MAINWIN_DEFAULT_FONT);

    if (!cfg.gentitle_format)
        cfg.gentitle_format = g_strdup("%p - %t");

    if (!cfg.outputplugin) {
#ifdef HAVE_OSS
        cfg.outputplugin =
            g_build_filename(PLUGIN_DIR, plugin_dir_list[0], "libOSS.so",
                             NULL);
#else
        /* FIXME: This implicitly means the output plugin that is
         * first in the alphabet will be used (usually the disk writer
         * plugin) */
        cfg.outputplugin = g_strdup("");
#endif
    }

    if (!cfg.eqpreset_default_file)
        cfg.eqpreset_default_file = g_strdup("dir_default.preset");

    if (!cfg.eqpreset_extension)
        cfg.eqpreset_extension = g_strdup("preset");
}


void
bmp_disabled_iplugins_foreach_func(gpointer key, gpointer value,
                                   gpointer data)
{
    gchar *tmp;

    if (GPOINTER_TO_INT(value) != 0)
        return;

    if (!disabled_iplugins) {
        tmp = g_strconcat(key, ":", NULL);
        disabled_iplugins = tmp;
        return;
    }

    tmp = g_strconcat(disabled_iplugins, key, ":", NULL);
    g_free(disabled_iplugins);
    disabled_iplugins = tmp;
}

void
bmp_config_save(void)
{
    GList *node;
    gchar *str;
    gint i;
    ConfigDb *db;

    g_hash_table_foreach(plugin_matrix,
                         (GHFunc) bmp_disabled_iplugins_foreach_func,
                         NULL);
    cfg.disabled_iplugins = g_strdup(disabled_iplugins);

    db = bmp_cfg_db_open();

    for (i = 0; i < ncfgbent; ++i)
        if (bmp_boolents[i].be_wrt)
            bmp_cfg_db_set_bool(db, NULL,
                                bmp_boolents[i].be_vname,
                                *bmp_boolents[i].be_vloc);

    for (i = 0; i < ncfgient; ++i)
        if (bmp_numents[i].ie_wrt)
            bmp_cfg_db_set_int(db, NULL,
                               bmp_numents[i].ie_vname,
                               *bmp_numents[i].ie_vloc);

    /* This is a bit lame .. it'll end up being written twice,
     * could do with being done a bit neater.  -larne   */
    bmp_cfg_db_set_int(db, NULL, "playlist_position",
                       playlist_get_position());

    for (i = 0; i < ncfgsent; ++i) {
        if (bmp_strents[i].se_wrt)
            bmp_cfg_db_set_string(db, NULL,
                                  bmp_strents[i].se_vname,
                                  *bmp_strents[i].se_vloc);
    }

    for (i = 0; i < 10; i++) {
        str = g_strdup_printf("equalizer_band%d", i);
        bmp_cfg_db_set_float(db, NULL, str, cfg.equalizer_bands[i]);
        g_free(str);
    }

    if (bmp_active_skin->path)
        bmp_cfg_db_set_string(db, NULL, "skin", bmp_active_skin->path);
    else
        bmp_cfg_db_unset_key(db, NULL, "skin");

    if (get_current_output_plugin())
        bmp_cfg_db_set_string(db, NULL, "output_plugin",
                              get_current_output_plugin()->filename);
    else
        bmp_cfg_db_unset_key(db, NULL, "output_plugin");

    str = general_stringify_enabled_list();
    if (str) {
        bmp_cfg_db_set_string(db, NULL, "enabled_gplugins", str);
        g_free(str);
    }
    else
        bmp_cfg_db_unset_key(db, NULL, "enabled_gplugins");

    str = vis_stringify_enabled_list();
    if (str) {
        bmp_cfg_db_set_string(db, NULL, "enabled_vplugins", str);
        g_free(str);
    }
    else
        bmp_cfg_db_unset_key(db, NULL, "enabled_vplugins");

    str = effect_stringify_enabled_list();
    if (str) {
        bmp_cfg_db_set_string(db, NULL, "enabled_eplugins", str);
        g_free(str);
    }
    else
        bmp_cfg_db_unset_key(db, NULL, "enabled_eplugins");

    if (cfg.filesel_path)
        bmp_cfg_db_set_string(db, NULL, "filesel_path", cfg.filesel_path);

    if (cfg.playlist_path)
        bmp_cfg_db_set_string(db, NULL, "playlist_path", cfg.playlist_path);

    bmp_cfg_db_set_int(db, NULL, "url_history_length",
                       g_list_length(cfg.url_history));

    for (node = cfg.url_history, i = 1; node; node = g_list_next(node), i++) {
        str = g_strdup_printf("url_history%d", i);
        bmp_cfg_db_set_string(db, NULL, str, node->data);
        g_free(str);
    }

    bmp_cfg_db_close(db);

    playlist_save(bmp_playlist_filename, FALSE);
}

static void
bmp_set_default_icon(void)
{
    GdkPixbuf *icon;

    icon = gdk_pixbuf_new_from_xpm_data((const gchar **) beep_mini_xpm);
    gtk_window_set_default_icon(icon);
    g_object_unref(icon);
}

static void
display_usage(void)
{
    g_print(_("Usage: beep-media-player [options] [files] ...\n\n"
              "Options:\n" "--------\n"));
    g_print("\n-h, --help             ");
    /* -h, --help switch */
    g_print(_("Display this text and exit"));
    g_print("\n-n, --session          ");
    /* -n, --session switch */
    g_print(_("Select BMP/XMMS session (Default: 0)"));
    g_print("\n-r, --rew              ");
    /* -r, --rew switch */
    g_print(_("Skip backwards in playlist"));
    g_print("\n-p, --play             ");
    /* -p, --play switch */
    g_print(_("Start playing current playlist"));
    g_print("\n-u, --pause            ");
    /* -u, --pause switch */
    g_print(_("Pause current song"));
    g_print("\n-s, --stop             ");
    /* -s, --stop switch */
    g_print(_("Stop current song"));
    g_print("\n-t, --play-pause       ");
    /* -t, --play-pause switch */
    g_print(_("Pause if playing, play otherwise"));
    g_print("\n-f, --fwd              ");
    /* -f, --fwd switch */
    g_print(_("Skip forward in playlist"));
    g_print("\n-e, --enqueue          ");
    /* -e, --enqueue switch */
    g_print(_("Don't clear the playlist"));
    g_print("\n-m, --show-main-window ");
    /* -m, --show-main-window switch */
    g_print(_("Show the main window"));
    g_print("\n-i, --sm-client-id     ");
    /* -i, --sm-client-id switch */
    g_print(_("Previous session ID"));
    g_print("\n-v, --version          ");
    /* -v, --version switch */
    g_print(_("Print version number and exit\n"));

#if 0
    g_print("\n-x, --xmms                   ");
    /* -x, --xmms switch */
    g_print(_("XMMS compatibility mode"));
    g_print("\n\n");
#endif

    exit(EXIT_SUCCESS);
}

static void
parse_cmd_line(gint argc, gchar ** argv, BmpCmdLineOpt * opt)
{
    static struct option long_options[] = {
        {"help", 0, NULL, 'h'},
        {"session", 1, NULL, 'n'},
        {"rew", 0, NULL, 'r'},
        {"play", 0, NULL, 'p'},
        {"pause", 0, NULL, 'u'},
        {"play-pause", 0, NULL, 't'},
        {"stop", 0, NULL, 's'},
        {"fwd", 0, NULL, 'f'},
        {"enqueue", 0, NULL, 'e'},
        {"show-main-window", 0, NULL, 'm'},
        {"version", 0, NULL, 'v'},
        {"sm-client-id", 1, NULL, 'i'},
        {"xmms", 0, NULL, 'x'},
        {0, 0, 0, 0}
    };

    gint c, i;
    gchar *filename;

    memset(opt, 0, sizeof(BmpCmdLineOpt));
    opt->session = -1;

    while ((c = getopt_long(argc, argv, "chn:rpusfemvtLS", long_options,
                            NULL)) != -1) {
        switch (c) {
        case 'h':
            display_usage();
            break;
        case 'n':
            opt->session = atoi(optarg);
            break;
        case 'r':
            opt->rew = TRUE;
            break;
        case 'p':
            opt->play = TRUE;
            break;
        case 'u':
            opt->pause = TRUE;
            break;
        case 's':
            opt->stop = TRUE;
            break;
        case 'f':
            opt->fwd = TRUE;
            break;
        case 't':
            opt->play_pause = TRUE;
            break;
        case 'm':
            opt->mainwin = TRUE;
            break;
        case 'e':
            opt->enqueue = TRUE;
            break;
        case 'v':
            dump_version();
            exit(EXIT_SUCCESS);
            break;
        case 'i':
            opt->previous_session_id = g_strdup(optarg);
            break;
        case 'c':
            opt->playcd = TRUE;
            break;
            /*  case 'x': */
            /*      opt->xmms_compat_mode = TRUE; */
            /*      break; */
        case 'S':
            opt->load_skins = TRUE;
            break;
        }
    }
    for (i = optind; i < argc; i++) {
        if (argv[i][0] == '/' || strstr(argv[i], "://"))
            filename = g_strdup(argv[i]);
        else {
            gchar *tmp = g_get_current_dir();
            filename = g_build_filename(tmp, argv[i], NULL);
            g_free(tmp);
        }
        opt->filenames = g_list_prepend(opt->filenames, filename);
    }
    opt->filenames = g_list_reverse(opt->filenames);
}

static void
handle_cmd_line_options(BmpCmdLineOpt * opt, gboolean remote)
{
    if (opt->session == -1) {
        if (!remote)
            opt->session = ctrlsocket_get_session_id();
        else
            opt->session = 0;
    }

    if (opt->filenames) {
        GList *node;
        gint pos = 0;

        if (opt->load_skins) {
            node = opt->filenames;
            xmms_remote_set_skin(opt->session, node->data);
            skin_install_skin(node->data);
            while (node) {
                g_free(node->data);
                node = g_list_next(node);
            }
            g_list_free(opt->filenames);
        }
        else {
            if (opt->enqueue && opt->play)
                pos = xmms_remote_get_playlist_length(opt->session);
            if (!opt->enqueue)
                xmms_remote_playlist_clear(opt->session);
            xmms_remote_playlist_add(opt->session, opt->filenames);
            node = opt->filenames;
            while (node) {
                g_free(node->data);
                node = g_list_next(node);
            }
            g_list_free(opt->filenames);
            if (opt->enqueue && opt->play &&
                xmms_remote_get_playlist_length(opt->session) > pos)
                xmms_remote_set_playlist_pos(opt->session, pos);
            if (!opt->enqueue)
                xmms_remote_play(opt->session);
        }
    }
    if (opt->rew)
        xmms_remote_playlist_prev(opt->session);
    if (opt->play)
        xmms_remote_play(opt->session);
    if (opt->pause)
        xmms_remote_pause(opt->session);
    if (opt->stop)
        xmms_remote_stop(opt->session);
    if (opt->fwd)
        xmms_remote_playlist_next(opt->session);
    if (opt->play_pause)
        xmms_remote_play_pause(opt->session);
    if (opt->mainwin)
        xmms_remote_main_win_toggle(opt->session, TRUE);
    if (opt->playcd)
        play_medium();
}

static void
segfault_handler(gint sig)
{
    g_printerr(_("\nReceived SIGSEGV\n\n"
                 "This could be a bug in BMP. If you don't know why this happened, "
                 "send a mail to us at beepmp-devel@lists.sourceforge.net\n\n"));
#ifdef HANDLE_SIGSEGV
    exit(EXIT_FAILURE);
#else
    abort();
#endif
}


#if 0
static gboolean
pposition_configure(GtkWidget * w, GdkEventConfigure * event, gpointer data)
{
    gint x, y;
    gdk_window_get_deskrelative_origin(w->window, &x, &y);
    if (x != 0 || y != 0) {
        g_warning("relative origin is at (%d, %d)", x, y);
        pposition_broken = TRUE;
    }
    gtk_widget_destroy(w);

    return FALSE;
}

static void
check_pposition(void)
{
    GtkWidget *window;
    GdkBitmap *mask;
    GdkGC *gc;
    GdkColor pattern;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    g_signal_connect(G_OBJECT(window), "configure_event",
                     G_CALLBACK(pposition_configure), NULL);
    gtk_widget_set_uposition(window, 0, 0);
    gtk_widget_realize(window);

    gtk_widget_set_usize(window, 1, 1);
    gdk_window_set_decorations(window->window, 0);

    mask = gdk_pixmap_new(window->window, 1, 1, 1);
    gc = gdk_gc_new(mask);
    pattern.pixel = 0;
    gdk_gc_set_foreground(gc, &pattern);
    gdk_draw_rectangle(mask, gc, TRUE, 0, 0, 1, 1);
    gdk_gc_destroy(gc);
    gtk_widget_shape_combine_mask(window, mask, 0, 0);

    gtk_widget_show(window);

    while (g_main_iteration(FALSE));
}
#endif

static void
bmp_setup_logger(void)
{
    if (!bmp_logger_start(bmp_log_filename))
        return;

    g_atexit(bmp_logger_stop);
}

static void
run_load_skin_error_dialog(const gchar * skin_path)
{
    const gchar *markup =
        N_("<b><big>Unable to load skin.</big></b>\n"
           "\n"
           "Check that skin at '%s' is usable and default skin is properly "
           "installed at '%s'\n");

    GtkWidget *dialog = gtk_message_dialog_new_with_markup(NULL,
                                                           GTK_DIALOG_MODAL,
                                                           GTK_MESSAGE_ERROR,
                                                           GTK_BUTTONS_CLOSE,
                                                           _(markup),
                                                           skin_path,
                                                           BMP_DEFAULT_SKIN_PATH);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

gint
main(gint argc, gchar ** argv)
{
    BmpCmdLineOpt options;

    /* Setup l10n early so we can print localized error messages */
    gtk_set_locale();
    bindtextdomain(PACKAGE, LOCALEDIR);
    bind_textdomain_codeset(PACKAGE, "UTF-8");
    textdomain(PACKAGE);

    bmp_init_paths();
    bmp_make_user_dir();
    bmp_setup_logger();

    /* Check GTK version. Really, this is only needed for binary
     * distribution since configure already checks. */
    if (!GTK_CHECK_VERSION(2, 4, 0)) {
        g_printerr(_
                   ("Sorry, your GTK+ version (%d.%d.%d) does not work with BMP.\n"
                    "Please use GTK+ %s or newer.\n"), gtk_major_version,
                   gtk_minor_version, gtk_micro_version, "2.4.0");
        exit(EXIT_FAILURE);
    }

    g_set_application_name(gettext(application_name));

    g_thread_init(NULL);
    if (!g_thread_supported()) {
        g_printerr(_("Sorry, threads isn't supported on your platform.\n\n"
                     "If you're on a libc5 based linux system and installed Glib & GTK+ before you\n"
                     "installed LinuxThreads you need to recompile Glib & GTK+.\n"));
        exit(EXIT_FAILURE);
    }

    gdk_threads_init();

    if (!gtk_init_check(&argc, &argv)) {
        if (argc < 2) {
            /* GTK check failed, and no arguments passed to indicate
               that user is intending to only remote control a running
               session */
            g_printerr(_("BMP: Unable to open display, exiting."));
            exit(EXIT_FAILURE);
        }

        handle_cmd_line_options(&options, TRUE);
        exit(EXIT_SUCCESS);
    }

#ifdef HAVE_GNOME_VFS
    if (!gnome_vfs_init()) {
        g_printerr(_("Could not initialize gnome-vfs.\n"));
        exit(EXIT_FAILURE);
    }

    g_atexit(gnome_vfs_shutdown);
#endif

    signal(SIGPIPE, SIG_IGN);   /* for controlsocket.c */
    signal(SIGSEGV, segfault_handler);

    parse_cmd_line(argc, argv, &options);

    g_random_set_seed(time(NULL));

    bmp_config_load();

    if (options.session != -1 || !ctrlsocket_setup()) {
        handle_cmd_line_options(&options, TRUE);
        exit(EXIT_SUCCESS);
    }

    check_wm_hints();
    //check_pposition();

    bmp_set_default_icon();

    mainwin_create();
    playlistwin_create();
    equalizerwin_create();

    if (!init_skins(cfg.skin)) {
        run_load_skin_error_dialog(cfg.skin);
        exit(EXIT_FAILURE);
    }

    GDK_THREADS_ENTER();

    init_plugins();
    read_volume(VOLSET_STARTUP);

    /* FIXME: these UI creation calls are delayed, because they deal
       directly with the plugin interface to set menu items */
    create_popups();

    playlist_load(bmp_playlist_filename);
    playlist_set_position(cfg.playlist_position);

    GDK_THREADS_LEAVE();

    ctrlsocket_start();

    handle_cmd_line_options(&options, FALSE);

    GDK_THREADS_ENTER();

    mainwin_set_info_text();

    if (cfg.player_visible)
        mainwin_show(TRUE);
    else if (!cfg.playlist_visible && !cfg.equalizer_visible)
        mainwin_show(TRUE);

    if (cfg.playlist_visible)
        playlistwin_show();

    if (cfg.equalizer_visible)
        equalizerwin_show(TRUE);

    playlist_start_get_info_thread();

    mainwin_attach_idle_func();

    create_prefs_window();

    /* FIXME: move this away, needed here because of metacity bug */
    hint_set_always(cfg.always_on_top);

    gtk_main();

    GDK_THREADS_LEAVE();

    return EXIT_SUCCESS;
}
